#include "common.h"
#include "Label.h"
#include "LabelSettings.h"
#include "Font.h"
#include "Texture.h"
#include "SystemInfo.h"

#include "ShowStyle.h"
#include "contextmenu.h"

#include <shellapi.h>
#include <shlobj.h>

#define TIMER_MOUSETRACK 1
#define TIMER_UPDATE 2
#define TIMER_SCROLL 3
#define TIMER_AUTOHIDE 4
#define TIMER_MOUSETRACK_MOVE 5
#define TIMER_ALPHAFADEIN 6
#define TIMER_ALPHAFADEOUT 7
#define TIMER_FRAMEANIMATION 8

extern LabelList labelList;
extern LabelList desktoplabelList;
extern LabelList shortcutlabelList;

Label::Label(const string &name):actualheight(0),actualwidth(0),actualx(0),actualy(0)
{
	this->name = name;

	hWnd = 0;
	hInstance = 0;
	box = 0;

	background = 0;
	font = 0;

	backgroundDC = 0;
	bufferDC = 0;
	backgroundBitmap = 0;
	bufferBitmap = 0;

	hoverActive = false;
	pressedActive = false;

	useHover = false;
	usePressed = false;
	
	mousePressed = false;
	mouseInside = false;

	movemodifierkeyPressed = false;
	moveButtonPressed = 0;

	visible = false;

	bUsingDefSkin = false;

	originalText.push_back("");
	current = 0;

	textChange = false;
	newlineCounter = 1;

	autoWidthMode = 0;
	autoHeightMode = 0;

	scroll = 0;
	scrollBackup = 1;
	scrollLimit = 0;
	scrollPosition = 0;
	frozen = false;

	alphaEnabled = false;
	
	//Animations
	frameCounter = 0;
	loopCounter = -1;
	normalAnim = false;
	hoverAnim = false;

	paintInProgress = false;
	reallyOverBlend = false;

	transparencyNeeded = true;

	//Plugins
	htmltext = false;
	maxHtmlTextWidth = 0;
	maxHtmlTextHeight = 0;

	anim = false;
}

Label::~Label()
{
	if (m_PluginHtmlText != NULL)
	{
		FreeLibrary(m_PluginHtmlText);
		m_PluginHtmlText = NULL;
	}
	if (m_PluginAnim != NULL)
	{
		FreeLibrary(m_PluginAnim);
		m_PluginAnim = NULL;
	}

	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
	}

	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_SCROLL);
		KillTimer(hWnd, TIMER_AUTOHIDE);
		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);
		KillTimer(hWnd, TIMER_FRAMEANIMATION);

		DestroyWindow(hWnd);
	}

	if(backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
	}
	if (backgroundDC)
		DeleteDC(backgroundDC);

	if(bufferBitmap)
	{
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		DeleteObject(bufferBitmap);
	}
	if (bufferDC)
		DeleteDC(bufferDC);

	if(!bUsingDefSkin)
		delete background;

	if(font != defaultSettings->font)
		delete font;
}

void Label::load(HINSTANCE hInstance, HWND box)
{
	this->hInstance = hInstance;
	
	if (IsWindow(box))
	{
		this->box = box;
	}
	else
	{
		this->box = NULL;
	}
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	hWnd = CreateWindowEx(box ? 0 : WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		box ? WS_CHILD : WS_POPUP,
		actualx, actualy,		
		actualwidth, actualheight,
		box ? box : GetLitestepWnd(), 
		0,
		hInstance,
		this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		//ToolTips for Labels
		if ( reconfigure() )
		{
			TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
										  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
										  NULL, NULL, hInstance, NULL);
			if (TooltipHints)
			{
				SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
							 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

				SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, tooltipmaxwidth);
			}

			strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());
			addHint(hWnd, tooltip);
		}

		repaint();
		if (autoWidthMode != 0 || autoHeightMode != 0)
			autoSize();
	}
	else
		delete this;
}

void Label::desktopload(HINSTANCE hInstance, string configLine)
{
	this->hInstance = hInstance;
	
    this->box = NULL;
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		WS_POPUP,
		actualx, actualy,			
		actualwidth, actualheight,
		GetLitestepWnd(),  
		0,
		hInstance,
		this);

	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);
		
		//ToolTips for Icons
		TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
									  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									  NULL, NULL, hInstance, NULL);
		if (TooltipHints)
		{
			SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

			SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 600);
		}

		desktopreconfigure(configLine);
		
		repaint();
	}
	else
		delete this;
}

void Label::shortcutload(HINSTANCE hInstance, string configLine, int lastx, int lasty)
{
	this->hInstance = hInstance;
	
    this->box = NULL;
	
	if (IsWindow(hWnd))	// window already exists...
		return;
		
	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabel",
		name.c_str(),
		WS_POPUP,
		actualx, actualy,		
		actualwidth, actualheight,	
		GetLitestepWnd(), 
		0,
		hInstance,
		this);

	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		//ToolTips for Shortcut2
		if ( shortcutreconfigure(configLine, lastx, lasty) )
		{	
			TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
									  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									  NULL, NULL, hInstance, NULL);
			
			if (TooltipHints)
				SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

			strcpy(tooltip, name.c_str());
			addHint(hWnd, tooltip);
		}
		
		repaint();
	}
	else
		delete this;
}

bool Label::reconfigure()
{
	LabelSettings settings(name.c_str());
	
	//Module Hook
	moduleHook = settings.moduleHook;

	//Normal Label = Type 0
	labelType = 0;

	//Transparency Settings
	transparencyMode = settings.transparencyMode;
	alphaTransparency = settings.alphaTransparency;
	if (alphaTransparency != 255 && box == NULL)
		alphaEnabled = true;
	alphaFade = settings.alphaFade;
	if (alphaFade && box == NULL)
		alphaEnabled = true;

	//Painting Background
	setBackground(settings.skin);	
	transparencyNeeded = background->isTransparencyNeeded();

	//AutoSize
	autoWidthMode = settings.autoWidthMode;
	autoHeightMode = settings.autoHeightMode;
	autoMinWidth = settings.width;
	autoMinHeight = settings.height;
	autoMaxWidth = settings.autoMaxWidth;
	autoMaxHeight = settings.autoMaxHeight;
	startCenterX = settings.x+(settings.width/2);
	startCenterY = settings.y+(settings.height/2);

	//Hover/Pressed Images
	if ( !(GetRCString(name.c_str(), "HoverImage", "")).empty() || !(GetRCString(name.c_str(), "HoverFrames", "")).empty())
		useHover = true;
	if ( !(GetRCString(name.c_str(), "PressedImage", "")).empty() )
		usePressed = true;

	//Animations
	overblendStyle = settings.overblendStyle;

	imageLoop = settings.imageLoop;
	imageFrameCount = settings.imageFrameCount;
	if (imageFrameCount != 0)
		normalAnim = true;
	imageFrameDelay = settings.imageFrameDelay;

	hoverimageLoop = settings.hoverimageLoop;
	hoverimageFrameCount = settings.hoverimageFrameCount;
	if (hoverimageFrameCount != 0)
		hoverAnim = true;
	hoverimageFrameDelay = settings.hoverimageFrameDelay;

	if (normalAnim)
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);

	//Painting Text
	setFont(settings.font);			
	setAlign(settings.align);
	setVertAlign(settings.vertAlign);
	lineBreak = settings.lineBreak;
	setLeftBorder(settings.leftBorder);
	setTopBorder(settings.topBorder);
	setRightBorder(settings.rightBorder);
	setBottomBorder(settings.bottomBorder);
	setText(settings.text);

	smooth = settings.smooth;
	
	//Scrolling
	scrollPadLength = settings.scrollPadLength;
	scrollInterval = settings.scrollInterval;
	scrollSpeed = settings.scrollSpeed;
	setScrolling(settings.scroll);

	//Updates
	alwaysUpdateContent = settings.alwaysUpdateContent;
	setUpdateInterval(settings.updateInterval);

	//Extra Settings
	labelGroup = settings.labelGroup;
	bUseFahrenheit = settings.bUseFahrenheit;

	//Events/Actions
	leftClickCommand = settings.leftClickCommand;
	leftDoubleClickCommand = settings.leftDoubleClickCommand;
	middleClickCommand = settings.middleClickCommand;
	middleDoubleClickCommand = settings.middleDoubleClickCommand;
	rightClickCommand = settings.rightClickCommand;
	rightDoubleClickCommand = settings.rightDoubleClickCommand;
	wheelDownCommand = settings.wheelDownCommand;
	wheelUpCommand = settings.wheelUpCommand;
	enterCommand = settings.enterCommand;
	leaveCommand = settings.leaveCommand;
	dropCommand = settings.dropCommand;
	textChangeCommand = settings.textChangeCommand;
	resizeCommand = settings.resizeCommand;

	//Event Regions
	labelLeftClickRegions = settings.labelLeftClickRegions;
	labelRightClickRegions = settings.labelRightClickRegions;
	labelMiddleClickRegions = settings.labelMiddleClickRegions;

	focusOnEnter = settings.focusOnEnter;

	//Moving
	moveable = settings.moveable;
	moveKey = settings.moveKey;
	moveButton = settings.moveButton;
	
	//Initial Repositioning
	if (box == NULL)
		reposition(settings.x, settings.y, settings.width, settings.height);
	else
	{
		RECT rc;
		GetClientRect(box, &rc);
		reposition(GetRCCoord(name, "X", 0, rc.right-rc.left),
					GetRCCoord(name, "Y", 0, rc.bottom-rc.top),
					GetRCDimen(name, "Width", 64, rc.right-rc.left),
					GetRCDimen(name, "Height", 64, rc.bottom-rc.top));
	}
		
	if (hWnd)
	{
		if(IsOS(OS_2KXP) && alphaEnabled)
		{
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);
			if (alphaFade)
				UpdateWindowTransparent(hWnd, 0);
			else
				UpdateWindowTransparent(hWnd, alphaTransparency);
		}

		setGhosted(settings.ghosted);

		//Plugins
		htmltext = settings.htmltext;
		if (htmltext)
		{
			if (!InitializeDLL("xhtmltext.dll"))
				htmltext = false;
		}
		anim = settings.anim;
		if (anim)
		{
			int tmpEffect = settings.animEffect;
			if (tmpEffect == -1)
				animEffect = Random;
			else if (tmpEffect == 0)
				animEffect = Spin;
			else if (tmpEffect == 1)
				animEffect = Vortex;
			else if (tmpEffect == 2)
				animEffect = ScatterGather;
			else if (tmpEffect == 3)
				animEffect = Spike;
			else if (tmpEffect == 4)
				animEffect = Fireworks;

			animSteps = settings.animSteps;
			animAfterSteps = settings.animAfterSteps;
			animDelay = settings.animDelay;
		    if (!InitializeDLL("xanim.dll"))
				anim = false;
		}

		//File Actions
		if (dropCommand.empty())
			DragAcceptFiles(hWnd, false);
		else
			DragAcceptFiles(hWnd, true);

		//AlwaysOnTop
		setAlwaysOnTop(settings.alwaysOnTop);	

		//Initial Show
		settings.startHidden ? hide() : show();
	}

	//Update Region (Magic Pink)
	if (transparencyMode == 1)
		updateRegion();

	//ToolTips
	originalTooltip = settings.tooltip;
	tooltipmaxwidth = settings.tooltipmaxwidth;
	if (!originalTooltip.empty())
		return true;
	else
		return false;
}

void Label::desktopreconfigure(string configLine)
{
	NameValuePair iconTextValues[] = {
		{ "top", 1 },
		{ "bottom", 2 },
		{ "left", 3 },
		{ "right", 4 },
		{ 0, 0 }
	};

	//DesktopLabel = Type 1
	labelType = 1;

	//Transparency Settings
	transparencyMode = 1;
	alphaTransparency = GetRCInt("xlabeldesktop", "AlphaTransparency", 255, 0 , 255);
	if (alphaTransparency != 255)
		alphaEnabled = true;
	alphaFade = false;

	//Painting Background
	overblendStyle = 0;

	//Prepare Settings
	//----------------------------------------------------------------
	char xname[32], ix[16], iy[16], path[MAX_PATH];
	char *buffers[] = {xname, ix, iy, path};

	LCTokenize(configLine.c_str(), buffers, 4, NULL);
	
	actualx = ParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN));
	actualy = ParseCoord(iy, 0, GetSystemMetrics(SM_CXSCREEN));
	actualwidth = GetRCInt("xlabeldesktop", "ItemWidth", 100);
	actualheight = GetRCInt("xlabeldesktop", "ItemHeight", 75);
	
	//----------------------------------------------------------------

	setBackground( GetDesktopTexture(path, this->getName()) );		

	//AutoSize
	autoWidthMode = 0;
	autoHeightMode = 0;

	//Hover/Pressed Images
	useHover = false;
	usePressed = false;

	//Animations
	normalAnim = false;
	hoverAnim = false;

	//Prepare Special DesktopSettings
	//----------------------------------------------------------------
	configSavePath = GetRCString("xlabeldesktop", "ConfigFile", "");
	rememberPosition = GetRCBoolean("xlabeldesktop", "RememberPosition", true);

	int iconSize = GetRCInt("xlabeldesktop", "IconSize", 32);
	int iconTextAlign = GetRCNamedValue("xlabeldesktop", "TextPosition", iconTextValues, 2);
	int iconTextSpacing = GetRCInt("xlabeldesktop" , "TextSpacing", 5);
	//----------------------------------------------------------------

	//Painting Text
	setFont( GetRCFont("xlabeldesktop") );
	setAlign(DT_CENTER);

	if (iconTextAlign == 1 || iconTextAlign == 2)
	{
		setLeftBorder(0);
		setRightBorder(0);
	}
	else if (iconTextAlign == 3)
	{
		setLeftBorder(0);
		setRightBorder(iconSize+iconTextSpacing+1);
	}
	else if (iconTextAlign == 4)
	{
		setLeftBorder(iconSize+iconTextSpacing+1);
		setRightBorder(0);
	}

	if (iconTextAlign == 1)
	{
		setTopBorder(0);
		setBottomBorder(iconSize+iconTextSpacing+1);
		setVertAlign(ALIGN_BOTTOM);
	}
	else if (iconTextAlign == 2)
	{
		setTopBorder(iconSize+iconTextSpacing+1);
		setBottomBorder(0);
		setVertAlign(ALIGN_TOP);
	}
	else if (iconTextAlign == 3 || iconTextAlign == 4)
	{
		setTopBorder(0);
		setBottomBorder(0);
		setVertAlign(ALIGN_CENTER);
	}

	lineBreak = true;

	iconCommand = path;
	bool itemIsFolder = false;
	
	HRESULT hres;
	IShellLink* psl;
	SHFILEINFO shFileInfo;

	char szGetPath[MAX_PATH];
	if (strstr(iconCommand.c_str(), "::{") != NULL)
	{
		itemIsFolder = true;
		SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME);
		setText( shFileInfo.szDisplayName );
		strcpy(tooltip , "Type:    ");
		strcat(tooltip , shFileInfo.szTypeName);
		strcat(tooltip, "\nTarget: ");
		strcat(tooltip , shFileInfo.szDisplayName);
	}
	else
	{
		SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME | SHGFI_ATTRIBUTES);
		setText( shFileInfo.szDisplayName );
		strcpy(tooltip , "Type:    ");
		strcat(tooltip , shFileInfo.szTypeName);
	
		if (shFileInfo.dwAttributes & SFGAO_FOLDER)
			itemIsFolder = true;
	}

	//Create the ShellLink object
	hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*) &psl);

	if (SUCCEEDED(hres))
	{
		IPersistFile* ppf;
		//Bind the ShellLink object to the Persistent File
		hres = psl->QueryInterface( IID_IPersistFile, (LPVOID *) &ppf);

		if (SUCCEEDED(hres))
		{
			WORD wsz[MAX_PATH];
			//Get a UNICODE wide string wsz from the Link path
			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, path, -1, wsz, MAX_PATH);
				
			//Read the link into the persistent file
			hres = ppf->Load(wsz, 0);
				
			if (SUCCEEDED(hres))
			{
				//Read the target information from the link object
				//UNC paths are supported (SLGP_UNCPRIORITY)
				psl->GetPath(szGetPath, MAX_PATH, NULL, SLGP_UNCPRIORITY);
					
				strcat(tooltip, "\nTarget: ");
				strcat(tooltip, szGetPath);

				psl->GetDescription(szGetPath, MAX_PATH);

				if (stricmp(szGetPath, "") != NULL)
				{
					strcat(tooltip, "\n\nDescription:\n");
					strcat(tooltip, szGetPath);

					SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 300);
				}
			}
			else if (strstr(path, "::{") == NULL)
			{
				strcat(tooltip, "\nName:   ");
				strcat(tooltip, iconCommand.c_str());
			}
		}
	}
	psl->Release();

	addHint(hWnd, tooltip);

	smooth = false;

	//Scrolling
	scrollPadLength = 10;
	scrollInterval = 100;
	scrollSpeed = 1;
	setScrolling(0);

	//Updates
	alwaysUpdateContent = false;
	setUpdateInterval(1000);

	//Extra Settings
	labelGroup = "xlabeldesktop";
	bUseFahrenheit = false;

	//Events/Actions
	leftClickCommand = "";
	if (itemIsFolder)
	{
		if (strstr(iconCommand.c_str(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}") != NULL)
		{
			dropCommand = "";
			itemIsFolder = false;
		}
		else if (strstr(iconCommand.c_str(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}") != NULL)
		{
			SHGetSpecialFolderPathA(NULL, path, CSIDL_PERSONAL, false);
			dropCommand = "copy \"";
			dropCommand.append( path );
			dropCommand.append( "\"" );
		}
		else if (strstr(iconCommand.c_str(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}") != NULL)
		{
			dropCommand = "";
			itemIsFolder = false;
		}
		else if (strstr(iconCommand.c_str(), "::{645FF040-5081-101B-9F08-00AA002F954E}") != NULL)
		{
			dropCommand = "delete files";
		}
		else
			dropCommand = ("copy \"" + iconCommand + "\"");

		leftDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
	}
	else
	{
		leftDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		dropCommand = "";
	}
	middleClickCommand = "!labelupdate xlabeldesktop";
	middleDoubleClickCommand = "";
	rightClickCommand = ".contextmenu";
	rightDoubleClickCommand = "";
	wheelDownCommand = "";
	wheelUpCommand = "";
	enterCommand = "";
	leaveCommand = "";
	textChangeCommand = "";
	resizeCommand = "";

	//Event Regions
	//labelLeftClickRegions = settings.labelLeftClickRegions;
	//labelRightClickRegions = settings.labelRightClickRegions;
	//labelMiddleClickRegions = settings.labelMiddleClickRegions;

	focusOnEnter = 0;

	//Moving
	moveable = true;
	moveKey = 1;
	moveButton = 1;

	//Initial Repositioning
	reposition(actualx, actualy , actualwidth, actualheight);
	
	if (hWnd)
	{
		if(IsOS(OS_2KXP) && alphaEnabled)
		{
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);
			UpdateWindowTransparent(hWnd, alphaTransparency);
		}

		setGhosted(false);

		//File Actions
		if (itemIsFolder)
			DragAcceptFiles(hWnd, true);
		else
			DragAcceptFiles(hWnd, false);

		//AlwaysOnTop
		setAlwaysOnTop(false);	

		//Initial Show
		GetRCBoolean("xlabeldesktop", "StartHidden") ? hide() : show();
	}

	tooltipmaxwidth = 300;

	//Update Region (Magic Pink)
	updateRegion();
}

bool Label::shortcutreconfigure(string configLine, int lastx, int lasty)
{
	//ShortcutLabel = Type 2
	labelType = 2;

	alphaTransparency = 255;
	alphaFade = false;

	overblendStyle = 0;

	//Prepare & Process Settings
	//----------------------------------------------------------------
	//Initializing
	LPSTR fLCommand = "";
    LPSTR fRCommand = "";
    LPSTR fMCommand = "";
	LPSTR fDropCommand = "";
	
	char name[32], ix[16], iy[16], normal[64], hover[64], pressed[64], rest[256];
	char *buffers[] = {name, ix, iy, normal, hover, pressed};

	LCTokenize(configLine.c_str(), buffers, 6, rest);

	//Hover/Pressed Images
	if ( stricmp(hover, ".none") != 0 )
		useHover = true;
	if ( stricmp(pressed, ".none") != 0 )
		usePressed = true;
	
	bool IsVisible;
	bool AcceptInput;
	bool Transparency;
	bool UseTooltip = GetRCBoolean("shortcut", "ShowCaptions", true);

	labelGroup = "0";
	char token[4096];
	LPCSTR nextToken = rest;
	bool hasCommand = true;

	if (GetToken(nextToken, token, &nextToken, false))
	{
		if (token[0] == '#')
		{
			//Lowercase
			strlwr(token);
			int tmpGroup = atoi(&token[1]);
			char buffer[16];
			labelGroup = itoa( tmpGroup, buffer, 10 );
			AcceptInput = !strchr(token, 'k');
            IsVisible = !strchr(token, 'h');
			alwaysOnTop = strchr(token, 't') != 0;
			Transparency = !strchr(token, 'o') != 0;
			moveable = strchr(token, 'm') != 0;
			rememberPosition = strchr(token, 's') != 0;

			if (UseTooltip)
				UseTooltip = !strchr(token, 'c');
			else
				UseTooltip = strchr(token, 'c') != 0;
          
            hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		}

		//Start Compatibility Modus to filter out sound definitions
		if (StrIPos(token, ".wav"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".none"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".def"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
	
		if (StrIPos(token, ".wav"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".none"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		else if (StrIPos(token, ".def"))
			hasCommand = GetToken(nextToken, token, &nextToken, false) != FALSE;
		//End Compatibility Modus to filter out sound definitions

        // ugly hack
        if (hasCommand)
        {
            char szCommands[4096];
            char szLCommand[4096] = { 0 };
            char szRCommand[4096] = { 0 };
            char szMCommand[4096] = { 0 };
            char szDropCommand[4096] = { 0 };
            
            if (nextToken)
            {
                sprintf(szCommands, "[%s] %s", token, nextToken);
            }
            else
            {
                sprintf(szCommands, "[%s]", token);
            }

            if (token[0] == '[')
            {
                LPSTR lpszCommands[] = { szLCommand, szRCommand, szMCommand, szDropCommand };
                CommandTokenize(szCommands, lpszCommands, 4, NULL);
            }
            else
            {
                // compatibility mode, leftclick only
                strcpy(szLCommand, szCommands);
            }
           
            if (szLCommand[0])
            {
                fLCommand = new char[strlen(szLCommand) + 1];
                strcpy(fLCommand, szLCommand);
            }

            if (szRCommand[0])
            {
                fRCommand = new char[strlen(szRCommand) + 1];
                strcpy(fRCommand, szRCommand);
            }

            if (szMCommand[0])
            {
                fMCommand = new char[strlen(szMCommand) + 1];
                strcpy(fMCommand, szMCommand);
            }

            if (szDropCommand[0])
            {
                fDropCommand = new char[strlen(szDropCommand) + 1];
                strcpy(fDropCommand, szDropCommand);
            }
            else if (szLCommand[0])
            {
                // default drag&drop command to leftclick command
                fDropCommand = new char[strlen(szLCommand) + 1];
                strcpy(fDropCommand, szLCommand);
            }
        }
	}

	//----------------------------------------------------------------

	configSavePath = GetRCString("Shortcut", "ConfigFile", "");	

	//Prepare "setBackground" Values
	//---------------------------------
	string images;
	images = normal;
	images.append(" ");
	images.append(hover);
	images.append(" ");
	images.append(pressed);

	string trans;
	if (Transparency)
	{
		transparencyMode = 1;
		trans = "true";
	}
	else
	{
		transparencyMode = 3;
		trans = "false";
	}
	//---------------------------------

	if (AcceptInput)
	{
		leftClickCommand = fLCommand;
		leftDoubleClickCommand = "";
		middleClickCommand = fMCommand;
		middleDoubleClickCommand = "";
		rightClickCommand = fRCommand;
		rightDoubleClickCommand = "";
		wheelDownCommand = "";
		wheelUpCommand = "";
		enterCommand = "";
		leaveCommand = "";
		dropCommand = "";
		textChangeCommand = "";
		resizeCommand = "";
	}
	else 
		setGhosted(true);

	//labelLeftClickRegions = settings.labelLeftClickRegions;
	//labelRightClickRegions = settings.labelRightClickRegions;
	//labelMiddleClickRegions = settings.labelMiddleClickRegions;

	focusOnEnter = 0;

	setScrolling(0);

	moveKey = 1;
	moveButton = 1;

	setBackground( GetShortcutTexture(images, trans) );	
	transparencyNeeded = background->isTransparencyNeeded();
	setFont( GetRCFont("") );
	
	HBITMAP	checkbitmap = LoadLSImage(normal, 0);
	GetLSBitmapSize(checkbitmap, &actualwidth, &actualheight);

	//Initial Repositioning
	reposition(ParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN), lastx), ParseCoord(iy, 0, GetSystemMetrics(SM_CYSCREEN), lasty), actualwidth, actualheight);	// FIXME? doesn't check old == new;

	if (hWnd)
	{
		//File Actions
		DragAcceptFiles(hWnd, FALSE);

		//AlwaysOnTop
		setAlwaysOnTop(alwaysOnTop);

		//Initial Show
		IsVisible ? show() : hide();
	}

	//Update Region (Magic Pink)
	if (transparencyMode == 1)
		updateRegion();

	//CleanUp
	if (fLCommand)
		delete[] fLCommand;
    if (fRCommand)
        delete[] fRCommand;
    if (fMCommand)
        delete[] fMCommand;
    if (fDropCommand)
        delete[] fDropCommand;

	return UseTooltip;
}

//------------------------------------------------------------
//Copy "LabelText" in mzscript variable
//------------------------------------------------------------
void Label::mzscriptvarcopy(string var)
{
	string bang = ("!varset " + var + " \"" + text + "\"");
	LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
}

//------------------------------------------------------------
//Copy "LabelPosition" in mzscript variables
//Export $..currentwidth$, $..currentheight$
//------------------------------------------------------------
void Label::posmzscriptvarcopy(string varx, string vary, bool setmzscript)
{
	string bang;
	string posx, posy;

	RECT cr;
	GetWindowRect(hWnd, &cr);

	POINT boxConverted;
	boxConverted.x = cr.left;
	boxConverted.y = cr.top;
	if (box != NULL)
		ScreenToClient(box, &boxConverted);

	char buffer[32];
	sprintf(buffer, "%d", boxConverted.x);
	posx = buffer;
	sprintf(buffer, "%d", boxConverted.y);
	posy = buffer;

	if (setmzscript)
	{
		bang = ("!varset " + varx + " \"" + posx + "\"");
		LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);

		bang = ("!varset " + vary + " \"" + posy + "\"");
		LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
	}
		
	PrefixedLSSetVariable(name, "CurrentX", posx);
	PrefixedLSSetVariable(name, "CurrentY", posy);
}

//------------------------------------------------------------
//Copy "LabelSize" in mzscript variables
//Export $..currentwidth$, $..currentheight$
//------------------------------------------------------------
void Label::sizemzscriptvarcopy(string varcx, string varcy, bool setmzscript)
{
	string bang;
	string cx, cy;

	RECT cr;
	GetWindowRect(hWnd, &cr);

	char buffer[32];
	sprintf(buffer, "%d", cr.right-cr.left);
	cx = buffer;
	sprintf(buffer, "%d", cr.bottom-cr.top);
	cy = buffer;

	if (setmzscript)
	{
		bang = ("!varset " + varcx + " \"" + cx + "\"");
		LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);

		bang = ("!varset " + varcy + " \"" + cy + "\"");
		LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
	}

	PrefixedLSSetVariable(name, "CurrentWidth", cx);
	PrefixedLSSetVariable(name, "CurrentHeight", cy);
}

//------------------------------------------------------------
//Set Background
//------------------------------------------------------------
void Label::setBackground(Texture *background)
{
	if(!bUsingDefSkin)
		delete this->background;

	this->background = background;

	bUsingDefSkin = (background == defaultSettings->skin);
}

//------------------------------------------------------------
//Set Parent Box (nearly never needed/called!)
//------------------------------------------------------------
void Label::setBox(HWND newparent)
{
	if (box == newparent)
		return;

	ModifyStyleEx(hWnd, WS_EX_LAYERED, 0);

	if (alwaysOnTop)
		setAlwaysOnTop(FALSE);
	
	box = newparent;
	
	ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
	SetParent(hWnd, newparent);
}

//------------------------------------------------------------
//"Modify Styles"
//------------------------------------------------------------
void Label::setAlwaysOnTop(bool alwaysOnTop)
{
	this->alwaysOnTop = alwaysOnTop;
	
	if (box)
		return;
	
	if(hWnd)
	{
		ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
		SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
		ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
		
		SetWindowPos(hWnd, alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
			0, 0, 0, 0,
			SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
	}
}

void Label::setAlpha(int factor)
{
	if (factor >= 0 && factor <= 255 && IsOS(OS_2KXP) && box == NULL)
	{
		if(!alphaEnabled)
		{
			alphaEnabled = true;
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);
		}

		UpdateWindowTransparent(hWnd, factor);
		alphaTransparency = factor;
	}
}

void Label::setGhosted(bool ghosted)
{
	this->ghosted = ghosted;
	
	if (box)
		return;
	
	if(hWnd)
	{
		if (ghosted)
			ModifyStyleEx(hWnd, 0, WS_EX_TRANSPARENT);
		else
			ModifyStyleEx(hWnd, WS_EX_TRANSPARENT, 0);
	}
}

//------------------------------------------------------------
//Set Font
//------------------------------------------------------------
void Label::setFont(Font *font)
{
	if(this->font != defaultSettings->font)
		delete this->font;

	this->font = font;
}

//------------------------------------------------------------
//Set Text Settings
//------------------------------------------------------------
void Label::setText(const string &text)
{
	originalText = split(text, ";");
		
	current = 0;
	update();
}

void Label::previous()
{
	current--;
	if(current < 0) current = originalText.size() - 1;
	update();
}

void Label::next()
{
	current++;
	if(current >= originalText.size()) current = 0;
	update();
}

void Label::setAlign(int align)
{
	if (this->align == align)
		return;

	this->align = align;
}

void Label::setVertAlign(int vertAlign)
{
	if (this->vertAlign == vertAlign)
		return;

	this->vertAlign = vertAlign;
}

void Label::setLeftBorder(int leftBorder)
{
	if (this->leftBorder == leftBorder)
		return;

	this->leftBorder = leftBorder;
}

void Label::setTopBorder(int topBorder)
{
	if (this->topBorder == topBorder)
		return;

	this->topBorder = topBorder;
}

void Label::setRightBorder(int rightBorder)
{
	if (this->rightBorder == rightBorder)
		return;

	this->rightBorder = rightBorder;
}

void Label::setBottomBorder(int bottomBorder)
{
	if (this->bottomBorder == bottomBorder)
		return;

	this->bottomBorder = bottomBorder;
}

void Label::setScrolling(int scrolling)
{
	if (scrolling != 0 && scroll == 0)
	{
		scrollPosition = 0;
		SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
	}
	else if (scrolling == 0 && scroll != 0)
	{
		KillTimer(hWnd, TIMER_SCROLL);
		scrollPosition = 0;
		repaint();
	}

	scrollLimit = 0;
	scroll = scrolling;
	if (scrolling != 0)
		scrollBackup = scrolling;
}

void Label::setScrollLimit(int limit)
{
	if (limit >= 0)
	{
		setScrolling(scrollBackup);
		scrollLimit = limit + 1;
	}
	else
		setScrolling(0);
}

void Label::setUpdateInterval(int updateInterval)
{
	this->updateInterval = updateInterval;
	if(hWnd && (dynamicText || dynamicTooltip || alwaysUpdateContent))
		SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
}

void Label::update()
{
	string oldText = text;

	//Tooltips
	if (!originalTooltip.empty() && TooltipHints)
	{
		strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());	

		if (dynamicTooltip)
			updateHint(hWnd, tooltip);
	}

	text = systemInfo->processLabelText(originalText[current], this, &dynamicText);
	
	//Start: Split in Multilines and check for longest line
	if(text != oldText)
	{
		vector<string> tmpLongest;
		int iter = 0;
		int longest = 0;
		string longestText;

		tmpLongest = split(text, "\n");
		longest = tmpLongest[iter].length();
		longestText = tmpLongest[iter];

		iter++;
		while (iter < tmpLongest.size())
		{
			if (longest < tmpLongest[iter].length())
			{
				longest = tmpLongest[iter].length();
				longestText = tmpLongest[iter];
			}
			iter++;
		}
		longestTextLine = longestText;
		newlineCounter = iter;

		textChange = true;
	}
	//End: Split in Multilines and check for longest line

	if(hWnd)
	{
		if(dynamicText || dynamicTooltip || alwaysUpdateContent)
			SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
		else
			KillTimer(hWnd, TIMER_UPDATE);
	}

	repaint();
	
	if(text != oldText)
	{
		if (autoWidthMode != 0 || autoHeightMode != 0)
			autoSize();
		if (transparencyMode == 1)
			updateRegion();
		if(!textChangeCommand.empty())
			LSExecute(NULL, textChangeCommand.c_str(), SW_SHOWNORMAL);
	}
}

//------------------------------------------------------------
//Set Animations
//------------------------------------------------------------
void Label::setAnimation(bool start, int loops)
{
	imageLoop = loops;
	normalAnim = start;

	if (start)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}
	else
	{
		frameCounter = 0;
		loopCounter = 0;
		KillTimer(hWnd, TIMER_FRAMEANIMATION);
		repaint(true);
	}	
}

//------------------------------------------------------------
//Move/Resize/Reposition
//------------------------------------------------------------
void Label::move(int newx, int newy, int steps, int time, bool moveBy)
{
	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		int counter = 0;
		int deltax = newx - actualx;
		int deltay = newy - actualy;

		bool growx;
		bool growy;
		if (deltax < 0)
			growx = false;
		else
			growx = true;

		if (deltay < 0)
			growy = false;
		else
			growy = true;

		div_t div_result = div( deltax, steps );
		deltax = div_result.quot;
		div_result = div( deltay, steps );
		deltay = div_result.quot;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;

		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		int dynx = actualx;
		int dyny = actualy;

		if(hWnd)
		{
			int stopper = 0;
			while (counter <= steps)
			{
				stopper = 0; // Stop if x and y finished, stopper = 2
				dynx = dynx+deltax;
				if (growx && dynx > newx)
				{
					dynx = newx;
					stopper++;
				}
				else if (!growx && dynx < newx)
				{
					dynx = newx;
					stopper++;
				}

				dyny = dyny+deltay;
				if (growy && dyny > newy)
				{
					dyny = newy;
					stopper++;
				}
				else if (!growy && dyny < newy)
				{
					dyny = newy;
					stopper++;
				}
				
				SetWindowPos(hWnd, 0, dynx, dyny, 0, 0,
					SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);

				SendMessage(hWnd, WM_PAINT, 0, 0);

				if (stopper == 2)
					counter = steps;
				
				counter++;

				Sleep(time);
			}

			actualx = newx;
			actualy = newy;

			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
					SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}
	else
	{
		if(hWnd)
		{
			if (moveBy)
			{
				newx = actualx + newx;
				newy = actualy + newy;
			}

			actualx = newx;
			actualy = newy;

			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
				SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = actualx+(actualwidth/2);
		startCenterY = actualy+(actualheight/2);
	}
}

void Label::resize(int newwidth, int newheight, int steps, int time, int sizeMode)
{
	POINT sizeHolder;
	sizeHolder.x = actualwidth;
	sizeHolder.y = actualheight;
	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		int counter = 0;
		int deltax = newwidth - actualwidth;
		int deltay = newheight - actualheight;

		bool growx;
		bool growy;
		if (deltax < 0)
			growx = false;
		else
			growx = true;

		if (deltay < 0)
			growy = false;
		else
			growy = true;

		div_t div_result = div( deltax, steps );
		deltax = div_result.quot;
		div_result = div( deltay, steps );
		deltay = div_result.quot;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;
		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		int dynwidth = actualwidth;
		int dynheight = actualheight;

		if(hWnd)
		{
			int stopper = 0;
			while (counter <= steps)
			{
				stopper = 0;  // Stop if width and height are finished, stopper = 2
				dynwidth = dynwidth+deltax;
				if (growx && dynwidth > newwidth)
				{
					dynwidth = newwidth;
					stopper++;
				}
				else if (!growx && dynwidth < newwidth)
				{
					dynwidth = newwidth;
					stopper++;
				}

				dynheight = dynheight+deltay;
				if (growy && dynheight > newheight)
				{
					dynheight = newheight;
					stopper++;
				}
				else if (!growy && dynheight < newheight)
				{
					dynheight = newheight;
					stopper++;
				}
				
				SetWindowPos(hWnd, 0, 0, 0, dynwidth, dynheight,
					SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
				
				SendMessage(hWnd, WM_PAINT, 0, 0);

				if (stopper == 2)
					counter = steps;

				counter++;

				Sleep(time);
			}

			actualheight = newheight;
			actualwidth = newwidth;

			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
					SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);
		
			//Force Repaint to get correct values for Resize Command (SetEvar must be called befire!)
			SendMessage(hWnd, WM_PAINT, 0, 0);
		}
	}
	else
	{
		if(hWnd)
		{
			if (sizeMode == -1)
			{
				if (newwidth > 0)
					newwidth = actualwidth - newwidth;
				else
					newwidth = actualwidth;
				if (newheight > 0)
					newheight = actualheight - newheight;
				else
					newheight = actualheight;
			}
			else if (sizeMode == 1)
			{
				if (newwidth > 0)
					newwidth = actualwidth + newwidth;
				else
					newwidth = actualwidth;
				if (newheight > 0)
					newheight = actualheight + newheight;
				else
					newheight = actualheight;
			}

			actualheight = newheight;
			actualwidth = newwidth;

			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
				SWP_NOMOVE | SWP_NOACTIVATE | SWP_NOZORDER);

			//Force Repaint to get correct values for Resize Command (SetEvar must be called befire!)
			SendMessage(hWnd, WM_PAINT, 0, 0);
		}
	}

	if (transparencyMode == 1)
		updateRegion();

	if(resizeCommand.length() > 0 && (actualwidth != sizeHolder.x || actualheight != sizeHolder.y))
		LSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);

	//Set new Rect for ToolTip Tracker
	if (TooltipHints)
		updateHint(hWnd, tooltip, true);
}

void Label::reposition(int newx, int newy, int newwidth, int newheight, int steps, int time)
{
	POINT sizeHolder;
	sizeHolder.x = actualwidth;
	sizeHolder.y = actualheight;
	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		int counter = 0;
		int deltax = newx - actualx;
		int deltay = newy - actualy;
		int deltasizex = newwidth - actualwidth;
		int deltasizey = newheight - actualheight;

		bool growx;
		bool growy;
		bool growsizex;
		bool growsizey;
		
		if (deltax < 0)
			growx = false;
		else
			growx = true;
		if (deltay < 0)
			growy = false;
		else
			growy = true;
		
		if (deltasizex < 0)
			growsizex = false;
		else
			growsizex = true;
		if (deltasizey < 0)
			growsizey = false;
		else
			growsizey = true;

		div_t div_result;
		div_result = div( deltax, steps );
		deltax = div_result.quot;
		div_result = div( deltay, steps );
		deltay = div_result.quot;

		div_result = div( deltasizex, steps );
		deltasizex = div_result.quot;
		div_result = div( deltasizey, steps );
		deltasizey = div_result.quot;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;
		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		if (deltasizex == 0 && growsizex)
			deltasizex = 1;
		else if (deltasizex == 0 && !growsizex)
			deltasizex = -1;
		if (deltasizey == 0 && growsizey)
			deltasizey = 1;
		else if (deltasizey == 0 && !growsizey)
			deltasizey = -1;

		int dynx = actualx;
		int dyny = actualy;
		int dynwidth = actualwidth;
		int dynheight = actualheight;

		if(hWnd)
		{
			int stopper = 0;
			while (counter <= steps)
			{
				stopper = 0; // Stop if x, y, width and height are finished, stopper = 4
				dynx = dynx+deltax;
				if (growx && dynx > newx)
				{
					dynx = newx;
					stopper++;
				}
				else if (!growx && dynx < newx)
				{
					dynx = newx;
					stopper++;
				}

				dyny = dyny+deltay;
				if (growy && dyny > newy)
				{
					dyny = newy;
					stopper++;
				}
				else if (!growy && dyny < newy)
				{
					dyny = newy;
					stopper++;
				}

				dynwidth = dynwidth+deltasizex;
				if (growsizex && dynwidth > newwidth)
				{
					dynwidth = newwidth;
					stopper++;
				}
				else if (!growsizex && dynwidth < newwidth)
				{
					dynwidth = newwidth;
					stopper++;
				}

				dynheight = dynheight+deltasizey;
				if (growsizey && dynheight > newheight)
				{
					dynheight = newheight;
					stopper++;
				}
				else if (!growsizey && dynheight < newheight)
				{
					dynheight = newheight;
					stopper++;
				}
				
				SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
					SWP_NOACTIVATE | SWP_NOZORDER);
				
				SendMessage(hWnd, WM_PAINT, 0, 0);

				if (stopper == 4)
					counter = steps;

				counter++;

				Sleep(time);
			}

			actualx = newx;
			actualy = newy;
			actualwidth = newwidth;
			actualheight = newheight;

			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
				SWP_NOACTIVATE | SWP_NOZORDER);

			//Force Repaint to get correct values for Resize Command (SetEvar must be called befire!)
			SendMessage(hWnd, WM_PAINT, 0, 0);
		}
	}
	else
	{
		if(hWnd)
		{
			actualx = newx;
			actualy = newy;
			actualheight = newheight;
			actualwidth = newwidth;

			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
				SWP_NOACTIVATE | SWP_NOZORDER);

			//Force Repaint to get correct values for Resize Command (SetEvar must be called befire!)
			SendMessage(hWnd, WM_PAINT, 0, 0);
		}
	}

	if (transparencyMode == 1)
		updateRegion();

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = actualx+(actualwidth/2);
		startCenterY = actualy+(actualheight/2);
	}

	if(resizeCommand.length() > 0 && (actualwidth != sizeHolder.x || actualheight != sizeHolder.y))
		LSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);

	//Set new Rect for ToolTip Tracker
	if (TooltipHints)
		updateHint(hWnd, tooltip, true);
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void Label::show()
{
	if (!hWnd)
		load(hInstance, box);
	
	if(hWnd)
	{
		KillTimer(hWnd, TIMER_AUTOHIDE);

		//Plugin Border Animation
		if (anim)
		{
			RECT r;
			GetWindowRect(hWnd, &r);
			if (pAnimShow != NULL)
				pAnimShow(r, animEffect, animSteps, animAfterSteps, animDelay);
		}
		if (alphaFade && IsOS(OS_2KXP) && box == NULL)
		{
			UpdateWindowTransparent(hWnd, 0);

			// Show it _first_
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);	

			// Redraw contents NOW - no flickering since the window's not visible
			RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW); 
				
			UpdateAlpha = 0;

			SetTimer(hWnd, TIMER_ALPHAFADEIN, 25, NULL);
		}
		else
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);

		visible = true;
		if (transparencyMode == 1)
			updateRegion();

		if (normalAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
		}
	}
}

void Label::showHide(double timeout)
{
	show();
	SetTimer(hWnd, TIMER_AUTOHIDE, timeout * 1000, NULL);
}

void Label::hide()
{
	if(hWnd)
	{
		if (alphaFade && IsOS(OS_2KXP) && box == NULL)
		{
			UpdateAlpha = alphaTransparency;

			SetTimer(hWnd, TIMER_ALPHAFADEOUT, 25, NULL);
		}
		else
		{
			ShowWindow(hWnd, SW_HIDE);
			if (anim)
			{
				RECT r;
				GetWindowRect(hWnd, &r);
				if (pAnimHide != NULL)
					pAnimHide(r, animEffect, animSteps, animAfterSteps, animDelay);
			}
		}

		visible = false;

		KillTimer(hWnd, TIMER_AUTOHIDE);

		if (normalAnim || hoverAnim)
			KillTimer(hWnd, TIMER_FRAMEANIMATION);
	}
}


//------------------------------------------------------------
//Clipboard Functions
//------------------------------------------------------------
void Label::clipboardCopy(string prefix)
{
	string toCopy = prefix.empty() ? text : (prefix + " " + text);
	SetClipboardText(hWnd, toCopy.c_str());
}

void Label::clipboardPaste(string prefix)
{
	char buffer[256];

	if(GetClipboardText(hWnd, buffer, 256))
	{
		string toPaste = prefix.empty() ? string(buffer) : (prefix + " " + string(buffer));
		setText(toPaste);
	}
}

//------------------------------------------------------------
//Region Events (Checking)
//------------------------------------------------------------
bool Label::checkRegion(int x, int y, StringList inputlist)
{
	bool found = false;
	if (!inputlist.empty())
	{
		for(StringListIterator it = inputlist.begin(); it != inputlist.end(); it++)
		{
			char left[16], top[16], right[16], bottom[16], command[128];
			char *buffers[] = {left, top, right, bottom};

			if(LCTokenize((*it).c_str(), buffers, 4, command) >= 4)
			{
				RECT rc;
				GetClientRect(hWnd, &rc);

				int intleft = ParseCoord(left, 0, rc.right-rc.left);
				int intright = ParseCoord(right, -0, rc.right-rc.left);
				int inttop = ParseCoord(top, 0, rc.bottom-rc.top);
				int intbottom  = ParseCoord(bottom, -0, rc.bottom-rc.top);

				if (intleft < intright && inttop < intbottom)
				{
					POINT pt;
					pt.x = x;
					pt.y = y;

					if(pt.x >= intleft && pt.x <= intright && pt.y >= inttop && pt.y <= intbottom)
					{
						if((*it).length() > 0)
						{	
							string temp = command;
							string bang = ("!execute [" + temp + "]");
							LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
							found = true;
							return found;
						}	
					}
				}
			}
		}
	}
	return found;
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

void Label::onLButtonDblClk(int x, int y)
{
	if(leftDoubleClickCommand.length() > 0)
		LSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onLButtonDown(int x, int y)
{
	//Silly hack to stop execution on mouseclick emulation
	//FocusOnEnter
	if (focusOnEnter != 2)
	{
	//--------------------------------------------------------------

	//Stop Animation Timer (Normal & Hover)
	KillTimer(hWnd, TIMER_FRAMEANIMATION);

	if(moveable && moveButton == 1)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 1;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		reallyOverBlend = true;
		repaint(true);
	
		if (transparencyMode == 1)
			updateRegion();
	}
	if (labelType == 2)
		ReorderZOrder();

	//--------------------------------------------------------------
	}
}

	
void Label::onLButtonUp(int x, int y)
{	
	//Silly hack to stop execution on mouseclick emulation
	//FocusOnEnter
	if (focusOnEnter != 2)
	{
	//--------------------------------------------------------------

	if(moveable && moveButton == 1)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelLeftClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);

				if(PtInRect(&rc, pt))
				{
					if(leftClickCommand.length() > 0)
						LSExecute(hWnd, leftClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			if ( rememberPosition )
			{
				string bang;
				bang = ("!labeldesktopsave " + iconCommand);
				LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
			}

			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			if ( rememberPosition )
			{
				string bang;
				bang = ("!labeldesktopsave " + iconCommand);
				LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);
			}

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelLeftClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);

			if(PtInRect(&rc, pt))
			{
				if(leftClickCommand.length() > 0)
					LSExecute(hWnd, leftClickCommand.c_str(), SW_SHOWNORMAL);
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		reallyOverBlend = true;
		repaint(true);

		if (transparencyMode == 1)
			updateRegion();
	}

	//Animation RESTART
	if (useHover)
	{
		//Restart Hover Anim
		if (hoverAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
	}
	//Restart Normal Anim
	else if (normalAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}

	if (labelType == 2)	
		ReorderZOrder();

	//--------------------------------------------------------------
	}
	else
		focusOnEnter = 1;
}

void Label::onMButtonDblClk(int x, int y)
{
	if(middleDoubleClickCommand.length() > 0)
		LSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onMButtonDown(int x, int y)
{
	//Stop Animation Timer (Normal & Hover)
	KillTimer(hWnd, TIMER_FRAMEANIMATION);

	if(moveable && moveButton == 3)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 3;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		reallyOverBlend = true;
		repaint(true);
	
		if (transparencyMode == 1)
			updateRegion();
	}
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onMButtonUp(int x, int y)
{
	if(moveable && moveButton == 3)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelMiddleClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);

				if(PtInRect(&rc, pt))
				{
					if(middleClickCommand.length() > 0)
						LSExecute(hWnd, middleClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelMiddleClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);

			if(PtInRect(&rc, pt))
			{
				if(middleClickCommand.length() > 0)
					LSExecute(hWnd, middleClickCommand.c_str(), SW_SHOWNORMAL);
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		reallyOverBlend = true;
		repaint(true);

		if (transparencyMode == 1)
			updateRegion();
	}

	//Animation RESTART
	if (useHover)
	{
		//Restart Hover Anim
		if (hoverAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
	}
	//Restart Normal Anim
	else if (normalAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}

	if (labelType == 2)
		ReorderZOrder();
}

void Label::onRButtonDblClk(int x, int y)
{
	if(rightDoubleClickCommand.length() > 0)
		LSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onRButtonDown(int x, int y)
{
	//Stop Animation Timer (Normal & Hover)
	KillTimer(hWnd, TIMER_FRAMEANIMATION);

	if(moveable && moveButton == 2)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			if (box)
				return;
			mousePressed = true;
			moveButtonPressed = 2;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		reallyOverBlend = true;
		repaint(true);
	
		if (transparencyMode == 1)
			updateRegion();
	}
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onRButtonUp(int x, int y)
{
	if(moveable && moveButton == 2)
	{
		if(mousePressed && !movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			
			if (!checkRegion(x, y, labelRightClickRegions))
			{
				POINT pt;
				pt.x = x;
				pt.y = y;

				RECT rc;
				GetClientRect(hWnd, &rc);

				if(PtInRect(&rc, pt))
				{
					if(rightClickCommand.length() > 0)
						LSExecute(hWnd, rightClickCommand.c_str(), SW_SHOWNORMAL);
				}
			}
			
			mousePressed = false;
			moveButtonPressed = 0;
		}
		else if (mousePressed && movemodifierkeyPressed)
		{
			//To be on the save side
			ReleaseCapture();
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

		if(mousePressed && !checkRegion(x, y, labelRightClickRegions))
		{
			POINT pt;
			pt.x = x;
			pt.y = y;

			RECT rc;
			GetClientRect(hWnd, &rc);

			if(PtInRect(&rc, pt))
			{
				if(rightClickCommand.length() > 0 && rightClickCommand != ".contextmenu")
					LSExecute(hWnd, rightClickCommand.c_str(), SW_SHOWNORMAL);
				else
					ContextMenu(iconCommand.c_str());
			}
		}
		mousePressed = false;
		moveButtonPressed = 0;
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		reallyOverBlend = true;
		repaint(true);

		if (transparencyMode == 1)
			updateRegion();
	}

	//Animation RESTART
	if (useHover)
	{
		//Restart Hover Anim
		if (hoverAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
	}
	//Restart Normal Anim
	else if (normalAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}

	if (labelType == 2)
		ReorderZOrder();
}

void Label::onWheelDown(int x, int y)
{
	if(wheelDownCommand.length() > 0)
		LSExecute(hWnd, wheelDownCommand.c_str(), SW_SHOWNORMAL);
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onWheelUp(int x, int y)
{
	if(wheelUpCommand.length() > 0)
		LSExecute(hWnd, wheelUpCommand.c_str(), SW_SHOWNORMAL);
	if (labelType == 2)
		ReorderZOrder();
}

void Label::onMouseEnter()
{
	if (focusOnEnter == 1)
	{	
		focusOnEnter = 2;
		mouse_event(MOUSEEVENTF_LEFTDOWN, NULL, NULL, NULL, NULL);
		mouse_event(MOUSEEVENTF_LEFTUP, NULL, NULL, NULL, NULL);
	}

	if (useHover)
	{
		//Kill AnimTimer from Normal Animation
		if (normalAnim)
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

		//If HoverAnim is set (hoverframecount != 0)
		if (hoverAnim)
		{
			hoverActive = true;
			pressedActive = false;
			frameCounter = 1;
			loopCounter = 0;
			//Start AnimTimer from Hover Animation
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
		//If HoverAnim is not set (hoverframecount == 0)
		else
		{
			frameCounter = 0;
			hoverActive = true;
			pressedActive = false;
			reallyOverBlend = true;
			repaint(true);

			if (transparencyMode == 1)
				updateRegion();
		}
	}	
	
	if(enterCommand.length() > 0)
		LSExecute(hWnd, enterCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onMouseLeave()
{
	if (useHover)
	{
		//If HoverAnim was active (hoverframecount != 0)
		if (hoverAnim)
		{
			hoverActive = false;
			pressedActive = false;
			frameCounter = 0;
			//Kill AnimTimer from Hover Animation
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				loopCounter = 0;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
			{
				reallyOverBlend = true;
				repaint(true);
				if (transparencyMode == 1)
					updateRegion();
			}
		}
		else
		{
			hoverActive = false;
			pressedActive = false;
			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
			{
				reallyOverBlend = true;
				repaint(true);
				if (transparencyMode == 1)
					updateRegion();
			}
		}
	}

	if(leaveCommand.length() > 0)
		LSExecute(hWnd, leaveCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onMouseMove(int x, int y)
{
	if(moveable)
	{
		if (mousePressed && movemodifierkeyPressed && moveButtonPressed == moveButton)
			SetTimer(hWnd, TIMER_MOUSETRACK_MOVE, 2, 0);
		else
		{
			KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);
			KillTimer(hWnd, TIMER_MOUSETRACK);

			if(!mouseInside)
			{
				if (alphaEnabled)
					SetFocus(hWnd);
				mouseInside = true;
				onMouseEnter();
				SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
			}
		}
	}
	else if (!moveable)
	{
		if(!mouseInside)
		{
			if (alphaEnabled)
				SetFocus(hWnd);
			mouseInside = true;
			onMouseEnter();
			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
		}
	}
}

//------------------------------------------------------------
//AutoWidthMode and AutoHeightMode Implementation
//------------------------------------------------------------
void Label::autoSize()
{
	RECT mainRect;
	GetWindowRect(hWnd, &mainRect);
	POINT boxConverted;

	boxConverted.x = mainRect.left;
	boxConverted.y = mainRect.top;
	if (box != NULL)
		ScreenToClient(box, &boxConverted);

	int width = mainRect.right - mainRect.left;
	int height = mainRect.bottom - mainRect.top;

	long textWidth;
	long textHeight;

	font->measure(bufferDC, longestTextLine, 0, &textWidth, &textHeight);

	if (htmltext && pMeasureHtmlParsedText != NULL)
	{
		RECT tmp;
		tmp.left = 0;
		tmp.right = 0;
		tmp.top = 0;
		tmp.bottom = 0;

		//POINT structure contains maximal width and height of HtmlText
		POINT sizeHtml = pMeasureHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, align, vertAlign, lineBreak);

		maxHtmlTextWidth = sizeHtml.x;
		maxHtmlTextHeight = sizeHtml.y;
	}

	int placeHolderWidth;
	int placeHolderHeight;

	if (!htmltext)
	{
		placeHolderWidth = textWidth+4;
		placeHolderHeight = newlineCounter*textHeight+4;
	}
	else if (htmltext)
	{
		placeHolderWidth = maxHtmlTextWidth+4;
		placeHolderHeight = maxHtmlTextHeight+4;
	}

	//For AUTOWIDTH/HEIGHT MODE	
	int newx;
	int newy;
	int newwidth;
	int newheight;

	int deltax;
	int deltay;

	if (autoWidthMode != 0)
		deltax = placeHolderWidth - (width - leftBorder - rightBorder);
	else
		deltax = 0;

	if (autoHeightMode != 0)
		deltay = placeHolderHeight - (height - topBorder - bottomBorder);
	else
		deltay = 0;

	newx = boxConverted.x;
	newy = boxConverted.y;
	newwidth = width + deltax;
	newheight = height + deltay;

	//Check MIN/MAX
	if (newwidth < autoMinWidth || newwidth > autoMaxWidth)
	{
		if (newwidth < autoMinWidth)
		{
			deltax = deltax + (autoMinWidth-newwidth);
			newwidth = autoMinWidth;
		}
		if (newwidth > autoMaxWidth)
		{
			deltax = deltax - (newwidth-autoMaxWidth);
			newwidth = autoMaxWidth;
		}
	}
	if (newheight < autoMinHeight || newheight > autoMaxHeight)
	{
		if (newheight < autoMinHeight)
		{
			deltay = deltay + (autoMinHeight-newheight);
			newheight = autoMinHeight;
		}
		if (newheight > autoMaxHeight)
		{
			deltay = deltay - (newheight-autoMaxHeight);
			newheight = autoMaxHeight;
		}
	}

	if ((autoWidthMode == 0 || autoWidthMode == 1) && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = boxConverted.x;
		newy = boxConverted.y;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 2)
	{
		newx = boxConverted.x;
		newy = boxConverted.y-deltay;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 3)
	{
		newx = boxConverted.x;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 2 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = boxConverted.x-deltax;
		newy = boxConverted.y;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 2)
	{
		newx = boxConverted.x-deltax;
		newy = boxConverted.y-deltay;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 3)
	{
		newx = boxConverted.x-deltax;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 3 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = startCenterX-(newwidth/2);
		newy = boxConverted.y;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 2)
	{
		newx = startCenterX-(newwidth/2);
		newy = boxConverted.y-deltay;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 3)
	{
		newx = startCenterX-(newwidth/2);
		newy = startCenterY-(newheight/2);
	}

	//Reposition to NEW Pos/Size ( ONLY IN AUTOWIDTH-, AUTOHEIGHT-MODE )
	reposition(newx, newy, newwidth, newheight);

	textChange = false;
}

void Label::repaint(bool invalidateCache)
{
	if(hWnd)
	{
		if(invalidateCache)
		{
			if(backgroundDC && backgroundBitmap)
			{
				backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
				DeleteObject(backgroundBitmap);
				backgroundBitmap = 0;
			}

			if(bufferDC && bufferBitmap)
			{
				bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
				DeleteObject(bufferBitmap);
				bufferBitmap = 0;
			}
		}

		InvalidateRect(hWnd, NULL, false);
	}
}

void Label::onPaint(HDC hDC)
{
	bool useFrames = false;
	RECT mainRect;
	GetWindowRect(hWnd, &mainRect);

	int width = mainRect.right - mainRect.left;
	int height = mainRect.bottom - mainRect.top;

	RECT completeFillRect;
	completeFillRect.left = 0;
	completeFillRect.top = 0;
	completeFillRect.right = width;
	completeFillRect.bottom = height;

	// keep a cached rendition of the background
	if(!backgroundBitmap)
	{
		if(!backgroundDC)
			backgroundDC = CreateCompatibleDC(hDC);

		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

		//True Transparency
		if (transparencyMode == 1)
		{
			HBRUSH hbrush;
			hbrush = CreateSolidBrush((RGB(255, 0, 255)));
			FillRect(backgroundDC, &completeFillRect, hbrush);
			DeleteObject(hbrush);
		}
		//PrePaint Background (Fake Transparency)
		else if(transparencyMode == 2)
		{
			if(box)
			{
				// save the previous DC contents as the background
				BitBlt(backgroundDC, 0, 0, width, height, hDC, 0, 0, SRCCOPY);
			}
			else
			{
				// paint desktop on display DC and then into background buffer
				PaintDesktopEx(backgroundDC, 0, 0, width, height, mainRect.left, mainRect.top, FALSE);
			}
		}
		//Default Background
		else if (transparencyMode == 3)
		{
			DrawEdge(backgroundDC,
					&completeFillRect,
					EDGE_RAISED,
					BF_RECT | BF_MIDDLE);
		}
		
		//Normal Background Painting if no Animation
		if ( (!normalAnim && !hoverAnim) || frameCounter == 0 || (usePressed && pressedActive))
		{
			if (hoverActive && useHover)
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 1);
			else if (pressedActive && usePressed)
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 2);
			else
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 0);
		}
		//Special Background Painting if Animation (normal or hover)
		else
		{
			useFrames = true;
			if (hoverActive && useHover)
				background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 1, frameCounter);
			else
				background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 0, frameCounter);
		}
	}

	// double buffer for flicker-free paint
	if(!bufferBitmap)
	{
		if(!bufferDC)
			bufferDC = CreateCompatibleDC(hDC);

		bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
	}

	// blt background into double buffer
	BitBlt(bufferDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

	//--------------------------------------------------------------------------------
	// start rendering text
	//--------------------------------------------------------------------------------
	//Create Region in which text drawing is allowed
	HRGN drawArea = CreateRectRgn( 0, 0, actualwidth, actualheight );
	HRGN allowedArea = CreateRectRgn( leftBorder, topBorder, width - rightBorder, height - bottomBorder );
	SelectClipRgn(bufferDC, allowedArea);
	DeleteObject(allowedArea);
	
	long textWidth;
	long textHeight;

	font->measure(bufferDC, longestTextLine, 0, &textWidth, &textHeight);

	if (htmltext && pMeasureHtmlParsedText != NULL)
	{
		RECT tmp;
		tmp.left = 0;
		tmp.right = 0;
		tmp.top = 0;
		tmp.bottom = 0;

		//POINT structure contains maximal width and height of HtmlText
		POINT sizeHtml = pMeasureHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, align, vertAlign, lineBreak);

		maxHtmlTextWidth = sizeHtml.x;
		maxHtmlTextHeight = sizeHtml.y;
	}

	int placeHolderWidth;
	int placeHolderHeight;

	if (!htmltext)
	{
		placeHolderWidth = textWidth;
		placeHolderHeight = newlineCounter*textHeight;
	}
	else if (htmltext)
	{
		placeHolderWidth = maxHtmlTextWidth;
		placeHolderHeight = maxHtmlTextHeight;
	}

	if ( ((scroll == 1 || scroll == 3) && placeHolderWidth > width - leftBorder - rightBorder) || ((scroll == 2 || scroll == 4) && placeHolderHeight > height - topBorder - bottomBorder) )
	{			
		if (scroll == 1 || scroll == 3)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = placeHolderWidth + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > placeHolderWidth + scrollPadLength )
				scrollPosition = 0;
		}
		else if (scroll == 2 || scroll == 4)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = placeHolderHeight + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > placeHolderHeight + scrollPadLength )
				scrollPosition = 0;
		}
		
		if (!htmltext)
		{
			if (scroll == 1 || scroll == 3)
			{
				font->apply(bufferDC,
						leftBorder,
						topBorder,
						scrollPosition - scrollPadLength,
						height - topBorder - bottomBorder,
						text,
						DT_RIGHT | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS ,
						vertAlign);

				font->apply(bufferDC,
						scrollPosition + leftBorder,
						topBorder,
						placeHolderWidth,
						height - topBorder - bottomBorder,
						text,
						DT_LEFT | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS ,
						vertAlign);
			}
			else if (scroll == 2 || scroll == 4)
			{
				font->apply(bufferDC,
					leftBorder,
					topBorder,
					width - leftBorder - rightBorder,
					scrollPosition - scrollPadLength,
					text,
					align | DT_NOPREFIX | DT_EXPANDTABS ,
					2);

				font->apply(bufferDC,
					leftBorder,
					scrollPosition + topBorder,
					width - leftBorder - rightBorder,
					placeHolderHeight,
					text,
					align | DT_NOPREFIX | DT_EXPANDTABS ,
					0);
			}
		}
		else if (htmltext && pDrawHtmlParsedText != NULL)
		{	
			if (scroll == 1 || scroll == 3)
			{
				RECT tmp;
						
				tmp.left = leftBorder;
				tmp.right = leftBorder + scrollPosition - scrollPadLength;
				tmp.top = topBorder;
				tmp.bottom = height - bottomBorder;
				
				pDrawHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, DT_RIGHT, vertAlign, false, smooth);
						
				tmp.left = leftBorder + scrollPosition;
				tmp.right = leftBorder + placeHolderWidth;
				tmp.top = topBorder;
				tmp.bottom = height - bottomBorder;
				
				pDrawHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, DT_LEFT, vertAlign, false, smooth);
			}
			else if (scroll == 2 || scroll == 4)
			{
				RECT tmp;
						
				tmp.left = leftBorder;
				tmp.right = width - rightBorder;
				tmp.top = scrollPosition;
				tmp.bottom = scrollPosition - scrollPadLength;
				
				pDrawHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, align, 2, false, smooth);
						
				tmp.left = leftBorder;
				tmp.right = width - rightBorder;
				tmp.top = scrollPosition - scrollPadLength - placeHolderHeight;
				tmp.bottom = placeHolderHeight;
				
				pDrawHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, align, 0, false, smooth);
			}
		}
	}
	else
	{
		if (!htmltext)
		{
			font->apply(bufferDC,
					leftBorder,
					topBorder,
					width - leftBorder - rightBorder,
					height - topBorder - bottomBorder,
					text,
					align | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS  | (lineBreak ? DT_WORDBREAK : 0),
					vertAlign);
		}
		else if (htmltext && pDrawHtmlParsedText != NULL)
		{	
			RECT tmp;
				
			tmp.left = leftBorder;
			tmp.right = width - rightBorder;
			tmp.top = topBorder;
			tmp.bottom = height - bottomBorder;
		
			pDrawHtmlParsedText(bufferDC, text.c_str(), font->GetLFont(), font->GetColor(), tmp, align, vertAlign, lineBreak, smooth);
		}
	}

	//Reset Region in which text drawing is allowed
	SelectClipRgn(bufferDC, drawArea);
	DeleteObject(drawArea);
	//--------------------------------------------------------------------------------
	// stop rendering text
	//--------------------------------------------------------------------------------

	//Normal Image Switch
	if (overblendStyle == 0 || useFrames || !reallyOverBlend)
	{
		// blt the double buffer to the display
		BitBlt(hDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);
	}
	//OverblendStyle Image Switch
	else if (overblendStyle == -1)
		MagicShow( bufferDC, 0, 0, hDC, 0, 0, width, height, 5, 2, rand()%20 );
	else
		MagicShow( bufferDC, 0, 0, hDC, 0, 0, width, height, 5, 2, overblendStyle );	

	//Set to idle ;) Only OverBlend on MouseOver or Pressed Events and back to normal
	reallyOverBlend = false;
}

void Label::updateRegion()
{
	if (transparencyNeeded)
	{
		HDC tmpDC = CreateCompatibleDC(bufferDC);
		HBITMAP tmpBmp = CreateBitmap(actualwidth, actualheight, 1, 32, NULL);
		HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

		//Draw Image
		if (bufferDC && bufferBitmap)
			BitBlt(tmpDC, 0, 0, actualwidth, actualheight, bufferDC, 0, 0, SRCCOPY);
		else
			onPaint(tmpDC);
			
		SelectObject(tmpDC, tmpObj);

		HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
		HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
		CombineRgn(windowRgn, region, NULL, RGN_COPY);

		if (!SetWindowRgn(hWnd, windowRgn, true))
			DeleteObject(windowRgn);	

		DeleteObject(region);
		DeleteObject(tmpBmp);
		DeleteObject(tmpObj);
		DeleteDC(tmpDC);
	}
}

void Label::onWindowPosChanged(WINDOWPOS *windowPos)
{
	repaint(true);
	
	RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW);

	if (labelType == 0)
		posmzscriptvarcopy("", "", false);
}

void Label::onSize(int width, int height)
{
	actualheight = height;
	actualwidth = width;

	repaint(true);

	RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW);

	if (labelType == 0)
		sizemzscriptvarcopy("", "", false);
}

void Label::relayMouseMessageToBox(UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	pt.x = GET_X_LPARAM(lParam);
	pt.y = GET_Y_LPARAM(lParam);

	if (box)
	{
		MapWindowPoints(hWnd, box, &pt, 1);
		PostMessage(box, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
	else
	{
		MapWindowPoints(hWnd, GetLitestepDesktop(), &pt, 1);
		PostMessage(GetLitestepDesktop(), message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
}

void Label::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:

			POINT pt;
			RECT rc;
			GetCursorPos(&pt);
			GetWindowRect(hWnd, &rc);

			if(!PtInRect(&rc, pt))
			{
				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;
				onMouseLeave();	
			}

		break;

		case TIMER_MOUSETRACK_MOVE:

			POINT newpt;
			GetCursorPos(&newpt);
			
			newpt.x = newpt.x-savedpt.x;
			newpt.y = newpt.y-savedpt.y;
			useHover = true;
			hoverActive = true;
			move(newpt.x, newpt.y, 0, 0);
		
		break;

		case TIMER_UPDATE:

			update();

		break;
		
		case TIMER_SCROLL:

			if (!frozen)
			{
				if (scroll == 1 || scroll == 2)
					scrollPosition -= scrollSpeed;
				else if (scroll == 3 || scroll == 4)
					scrollPosition += scrollSpeed;
				repaint();
			}

		break;

		case TIMER_AUTOHIDE:

			hide();

		break;

		case TIMER_ALPHAFADEIN:

			if (UpdateAlpha < alphaTransparency)
			{
				// Raise the alpha value (towards alphatransparency) and update the window.
				UpdateAlpha += 5;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
			}
			else
			{
				// Reached maximum alpha. Kill the timer
				UpdateAlpha = alphaTransparency;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
			}

		break;

		case TIMER_ALPHAFADEOUT:

			if (UpdateAlpha >= 5)
			{
				// Reduce the alpha value (towards transparency) and update the window.
				UpdateAlpha -= 5;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
			}
			else
			{
				// Reached minimum alpha. Kill the timer
				UpdateAlpha = 0;
				UpdateWindowTransparent(hWnd, UpdateAlpha);
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				ShowWindow(hWnd, SW_HIDE);
				if (anim)
				{
					RECT r;
					GetWindowRect(hWnd, &r);
					if (pAnimHide != NULL)
						pAnimHide(r, animEffect, animSteps, animAfterSteps, animDelay);
				}
			}

		break;

		case TIMER_FRAMEANIMATION:

			if (hoverActive && useHover)
			{
				if (frameCounter == hoverimageFrameCount)
				{
					frameCounter = 0;
					loopCounter++;
				}
				if (loopCounter == hoverimageLoop)
				{
					frameCounter = -1;
					loopCounter = 0;
					KillTimer(hWnd, TIMER_FRAMEANIMATION);
				}
				
			}
			else
			{
				if (frameCounter == imageFrameCount)
				{
					frameCounter = 0;
					loopCounter++;
				}
				if (loopCounter == imageLoop)
				{
					frameCounter = -1;
					loopCounter = 0;
					KillTimer(hWnd, TIMER_FRAMEANIMATION);
				}
			}
			frameCounter++;
			repaint(true);

		break;
	}
}

bool Label::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	extern HWND messageHandler;
	
	if (labelType == 0 || labelType == 2)
	{
		if (moveKey == 3)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT));
		else if (moveKey == 2)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_SHIFT));
		else if (moveKey == 1)
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
		else 
			movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
	}
	else if (labelType == 1)
		movemodifierkeyPressed = true;

	if (moveable && !movemodifierkeyPressed)
		KillTimer(hWnd, TIMER_MOUSETRACK_MOVE);

	switch(message)
	{
		case LM_SETLABELTEXT:
		{
			setText(string((const char *) lParam));
			return true;
		}

		case WM_ACTIVATE:
		{
			if (labelType == 1)
			{
				if (wParam == WA_INACTIVE)
				{
					useHover = false;
					hoverActive = false;
					repaint(true);
				}
				else if (wParam == WA_CLICKACTIVE)
				{
					useHover = true;
					hoverActive = true;
					repaint(true);
					SendMessage(hWnd, WM_PAINT, 0, 0);
				}
			}
			return true;
		}

		case WM_COPYDATA:
		{
			COPYDATASTRUCT *cds = (COPYDATASTRUCT *) lParam;
			
			if(cds->dwData == LM_SETLABELTEXT)
				setText(string((const char *) cds->lpData));

			return false;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DESTROY:
		{
			if(box)		// FIXME: this is the old workaround
			{
				box = NULL;
				hide();
			}
				
			hWnd = NULL;	// FIXME: dirty hack
			
			return false;
		}

		case WM_DROPFILES:
		{
			if(dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			int intFileAction;

			char action[32], szTarget[MAX_PATH];;
			char *buffers[] = {action, szTarget};

			LCTokenize(dropCommand.c_str(), buffers, 2, NULL);

			if (stricmp(action, "move") == 0)
				intFileAction = 1;
			else if (stricmp(action, "copy") == 0)
				intFileAction = 2;
			else if (stricmp(action, "delete") == 0)
				intFileAction = 0;
			else
				intFileAction = -1;

			if (intFileAction >= 0)
			{
				char *szSourceFiles = NULL;
				char buffer[MAX_PATH];
				int numFiles;
				UINT len,scursor = 0;

				numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

				szSourceFiles = new char[MAX_PATH*numFiles+1];
				scursor = 0;

				for (int i=0;i<numFiles;i++)
				{
					len = DragQueryFile((HDROP) wParam, i, buffer, MAX_PATH);
					
					strcpy(szSourceFiles+scursor,buffer);
					scursor += len + 1;
				}
					
				DragFinish((HDROP) wParam);

				string check = szTarget;
				if (check.substr(1, check.length()-1) == "\\")
					check.append( "*" );
				else
					check.append( "\\*" );
				
				//Test Path for Existence
				WIN32_FIND_DATA FindData;
				HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

				if(hFindHandle != INVALID_HANDLE_VALUE || intFileAction == 0)
				{
					szSourceFiles[scursor] = 0;
					strcpy(buffer, szTarget);
					buffer[strlen(buffer)+1] = 0;
					
					SHFILEOPSTRUCT shfop;
					shfop.hwnd = NULL;
					if (intFileAction == 1)
						shfop.wFunc = FO_MOVE;
					else if (intFileAction == 2)
						shfop.wFunc = FO_COPY;
					else if (intFileAction == 0)
						shfop.wFunc = FO_DELETE;
					shfop.pFrom = szSourceFiles;
					shfop.pTo = buffer;
					shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
					SHFileOperation(&shfop);
				}
				
				delete[] szSourceFiles;
			}
			else if (intFileAction == -1)
			{
				int numDropped, i;
				char szFile[MAX_PATH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_PATH);
					if (szFile && szFile[0])
					{
						string fullCommand = dropCommand + " \"" + szFile + "\""; // FIXME: ugly hack
						LSExecute(hWnd, fullCommand.c_str(), SW_SHOWNORMAL);
					}
				}
				DragFinish((HDROP) wParam);
			}
			
			return 0;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			if(leftDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			if(leftDoubleClickCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && !moveable && focusOnEnter == 0)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			if(leftDoubleClickCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && !moveable && focusOnEnter == 0)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			if(middleDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			if(middleDoubleClickCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			if(middleDoubleClickCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			if(rightDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			if(rightDoubleClickCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			if(rightDoubleClickCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && !moveable)
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MOUSEWHEEL:
		{
			if ((short)(HIWORD(wParam)) < 0)
			{
				if(wheelDownCommand.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onWheelDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
			else
			{
				if(wheelUpCommand.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}
				
				onWheelUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
		}

		case WM_MOUSEMOVE:
		{
			onMouseMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_PAINT:
		{
			//Only Start New PAINT if old is finished ;)
			if (!paintInProgress || overblendStyle == 0)
			{
				paintInProgress = true;

				PAINTSTRUCT ps;
				HDC hDC;

				if(!wParam)
					hDC = BeginPaint(hWnd, &ps);
				else
					hDC = (HDC) wParam;

				onPaint(hDC);

				if(!wParam)
					EndPaint(hWnd, &ps);

				paintInProgress = false;
			}

			return true;
		}

		case WM_SIZE:
		{
			onSize((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case WM_WINDOWPOSCHANGED:
		{
			onWindowPosChanged((WINDOWPOS *) lParam);
			return false;
		}
	}

	return false;
}

LRESULT Label::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Label *label = NULL;

	if(message == WM_NCCREATE)
	{
		label = (Label *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		label->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) label);
	}
	else
		label = (Label *) GetWindowLong(hWnd, 0);

	if(label)
	{
		LRESULT lResult = 0;

		if(label->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//------------------------------------------------------------
//Tooltips (Add, Update, Remove)
//------------------------------------------------------------

void Label::addHint(HWND hWnd, LPSTR caption)
{
	TOOLINFO ti;
	RECT clientRect;

	if (!TooltipHints)
		return ;
	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_ADDTOOL, 0, LPARAM(&ti));
}

void Label::updateHint(HWND hWnd, LPSTR caption, bool updatesize)
{
	TOOLINFO ti;
	RECT clientRect;

	if (!TooltipHints)
		return;
	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_UPDATETIPTEXT, 0, LPARAM(&ti));
	if (updatesize)
		SendMessage(TooltipHints, TTM_NEWTOOLRECT, 0, LPARAM(&ti));
}

void Label::removeHint(HWND hWnd)
{
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	if (!TooltipHints)
		return ;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_DELTOOL, 0, LPARAM(&ti));
}

//------------------------------------------------------------
//Initialize and Update AlphaTransparency
//------------------------------------------------------------
BOOL Label::UpdateWindowTransparent(HWND hWnd, unsigned char factor)
{
   /* First, see if we can get the API call we need. If we've tried
    * once, we don't need to try again. */
   if (!initialized)
   {
		HMODULE hDLL = LoadLibrary ("user32");

		pSetLayeredWindowAttributes = 
		  (PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

		//pUpdateLayeredWindow = 
		//	(PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

		initialized = TRUE;
   
	   if (pSetLayeredWindowAttributes == NULL /*|| pUpdateLayeredWindow == NULL*/)
		  return FALSE;
  
	   SetLastError(0);

	   if (GetLastError())
		  return FALSE;
	}

   /* Now, we need to set the 'layered window attributes'. This
    * is where the alpha values get set. */
  
	return pSetLayeredWindowAttributes (hWnd, 0, factor, LWA_ALPHA);
}

//------------------------------------------------------------
//Reorder "ShortcutLabels" to keep compatible to sc2
//------------------------------------------------------------
void Label::ReorderZOrder()
{
	for(LabelListIterator iter = shortcutlabelList.begin(); iter != shortcutlabelList.end(); iter++)
	{
		if((*iter)->getName() == this->getName())
		{
			iter++;
			if (iter != NULL)
			{
				SetWindowPos(hWnd, (*iter)->hWnd,
					0, 0, 0, 0,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			}
			break;
		}
	}
}

//------------------------------------------------------------
//Initialize xLabel Plugins
//------------------------------------------------------------
bool Label::InitializeDLL(string pluginName)
{
	//Load Plugin DLL's
	if (stricmp(pluginName.c_str(), "xhtmltext.dll") == 0)
	{
		m_PluginHtmlText = LoadLibrary(pluginName.c_str());
		
		if(m_PluginHtmlText == NULL)
		{
			char tmpSz[4096];
			// Try to load from xLabel.dll's folder
			string currentBuild = V_NAME;
			currentBuild.append("-");
			currentBuild.append(V_VERSION);
			currentBuild.append(".dll");

			GetModuleFileName(GetModuleHandle(currentBuild.c_str()), tmpSz, 4096);
			tmpSz[strlen(tmpSz) - currentBuild.length()] = 0;	// Remove the DLL name
			strcat(tmpSz, pluginName.c_str());

			m_PluginHtmlText = LoadLibrary(tmpSz);

			if(m_PluginHtmlText == NULL)
			{
				pluginName = "Plugin not found!\n\n";
				pluginName.append(tmpSz);
				pluginName.append("\n\nPlease check Path and PluginName!\nIf it still refuses to work, copy xhtmltext.dll into your LitestepDir!");
				MessageBox(NULL, pluginName.c_str(), "xLabel Plugin Informations", MB_SETFOREGROUND);
				return false;
			}
		}

		if (m_PluginHtmlText != NULL)
		{
			pDrawHtmlParsedText = (PDHPT)GetProcAddress(m_PluginHtmlText, "DrawHtmlParsedText");
			pMeasureHtmlParsedText = (PMHPT)GetProcAddress(m_PluginHtmlText, "MeasureHtmlParsedText");
		
			if (pDrawHtmlParsedText == NULL || pMeasureHtmlParsedText == NULL)
			{
				MessageBox(NULL, "Couldn't initialise PluginFunctions!\n\nMaybe you use an older xhtmltext.dll version, than xlabel.dll version!", "xLabel Plugin Informations", MB_SETFOREGROUND);
				return false;
			}
			else
				return true;
		}
	}
	else if (stricmp(pluginName.c_str(), "xanim.dll") == 0)
	{
		m_PluginAnim = LoadLibrary(pluginName.c_str());

		if(m_PluginAnim == NULL)
		{
			char tmpSz[4096];
			// Try to load from xLabel.dll's folder
			string currentBuild = V_NAME;
			currentBuild.append("-");
			currentBuild.append(V_VERSION);
			currentBuild.append(".dll");

			GetModuleFileName(GetModuleHandle(currentBuild.c_str()), tmpSz, 4096);
			tmpSz[strlen(tmpSz) - currentBuild.length()] = 0;	// Remove the DLL name
			strcat(tmpSz, pluginName.c_str());
			m_PluginAnim = LoadLibrary(tmpSz);

			if(m_PluginAnim == NULL)
			{
				pluginName = "Plugin not found!\n\n";
				pluginName.append(tmpSz);
				pluginName.append("\n\nPlease check Path and PluginName!\nIf it still refuses to work, copy xanim.dll into your LitestepDir!");
				MessageBox(NULL, pluginName.c_str(), "xLabel Plugin Informations", MB_SETFOREGROUND);
				return false;
			}
		}

		if (m_PluginAnim != NULL)
		{
			pAnimShow = (PAS)GetProcAddress(m_PluginAnim, "AnimEffectShow");
			pAnimHide = (PAH)GetProcAddress(m_PluginAnim, "AnimEffectHide");
		
			if (pAnimShow == NULL || pAnimHide == NULL)
			{
				MessageBox(NULL, "Couldn't initialise PluginFunctions!\nMaybe you use an older xanim.dll version, than xlabel.dll version!", "xLabel Plugin Informations", MB_SETFOREGROUND);
				return false;
			}
			else
				return true;
		}
	}

	return false;
}